#include "TecModule.h"
#include "TecParameters.h"

/**********************************************************************************************************************/

TecModule::TecModule(EntityOwner* owner, const std::string& name, const std::string& description,
    ctk::HierarchyModifier hierarchyModifier, const std::unordered_set<std::string>& tags)
: ctk::ApplicationModule(owner, name, description, hierarchyModifier, tags) {
  // example how to fill the map
    writeIntMap.emplace(
        TecFrame::TecParameter::ExpFanControlEnable, ChimeraTK::ScalarPollInput<int>{this, "ctrl_ch1_fanControlEnable_wr", "unit", "description"}); 
      
    writeFloatMap.emplace(
        TecFrame::TecParameter::TemTargetObjectTemp, ChimeraTK::ScalarPollInput<float>{this, "ctrl_ch1_targetObjectTemp_wr", "unit", "description"});        
        
    readIntMap.emplace(
        TecFrame::TecParameter::ExpFanControlEnable, ChimeraTK::ScalarOutput<int>{this, "ctrl_ch1_fanControlEnable_rd", "unit", "description"});       
    readIntMap.emplace(
        TecFrame::TecParameter::MonSerialNumber, ChimeraTK::ScalarOutput<int>{this, "mon_serialNumber_rd", "unit", "description"});   
           
    readFloatMap.emplace(
        TecFrame::TecParameter::TemTargetObjectTemp, ChimeraTK::ScalarOutput<float>{this, "ctrl_ch1_targetObjectTemp_rd", "unit", "description"});         
    readFloatMap.emplace(
        TecFrame::TecParameter::MonSinkTemperature, ChimeraTK::ScalarOutput<float>{this, "mon_ch1_SinkTemp_rd", "unit", "description"});      
        
      
  // Note: The maps writeGroupMap, writeGroupToFlashMap, readGroupMap need to be filled in the mainLoop(). The
  // corresponding ScalarPushInputs need to be created here and placed onto the groupCommandMap.

  // example how to fill a group command -> this part needs to later automatically generated from std::list<ParameterDescriptor> descriptors
  // there is ticket to automate it
  // for now manually entries prepared for start of the Focus Week

  // commands for read-only for channel 1 parameters in group 1 where TecFrame::TecParameter::MonSinkTemperature query includes
  groupCommandMap.emplace("readCh1Group1",
      ChimeraTK::ScalarPushInput<int>{this, "readCh1Group1_wr", "", "Trigger read of channel 1 read-only parameters group 1"});
  // read-only for generic parameter in group 1 for TecFrame::TecParameter::MonSerialNumber query includes
  groupCommandMap.emplace("readGroup1",
      ChimeraTK::ScalarPushInput<int>{this, "readGroup1_wr", "", "Trigger read of generic read-only parameters group 1."});
  
  // commands for read-write group 7 for channel 1 where TecFrame::TecParameter::ExpFanControlEnable query includes
  groupCommandMap.emplace("readCh1WriteGroup7",
      ChimeraTK::ScalarPushInput<int>{this, "readCh1WriteGroup7_wr", "", "Trigger read of channel 1 read-write group 7"});
  groupCommandMap.emplace("writeCh1Group7",
      ChimeraTK::ScalarPushInput<int>{this, "writeCh1Group7_wr", "", "Trigger write of channel 1 read-write group 7"});
  groupCommandMap.emplace("writeToFlashCh1Group7",
      ChimeraTK::ScalarPushInput<int>{this, "writeToFlashCh1Group7_wr", "", "Trigger write to FLASH of channel 1 read-write group 7"});
  groupCommandMap.emplace("setRbCh1Group7",
      ChimeraTK::ScalarPushInput<int>{this, "setRbCh1Group7_wr", "", "Trigger set from readback for channel 1 read-write group 7"});      
      
  // commands for read-write group 1 for channel 1 where TecFrame::TecParameter::TemTargetObjectTemp query includes
  groupCommandMap.emplace("readCh1WriteGroup1",
      ChimeraTK::ScalarPushInput<int>{this, "readCh1WriteGroup1_wr", "", "Trigger read of channel 1 read-write group 1"});
  groupCommandMap.emplace("writeCh1Group1",
      ChimeraTK::ScalarPushInput<int>{this, "writeCh1Group1_wr", "", "Trigger write of channel 1 read-write group 1"});
  groupCommandMap.emplace("writeToFlashCh1Group1",
      ChimeraTK::ScalarPushInput<int>{this, "writeToFlashCh1Group1_wr", "", "Trigger write to FLASH of channel 1 read-write group 1"});
  groupCommandMap.emplace("setRbCh1Group1",
      ChimeraTK::ScalarPushInput<int>{
          this, "setRbCh1Group1_wr", "", "Trigger set from readback for channel 1 read-write group 1"});

  // TODO: Should this be combined with the implementation of #6567?
  for(auto& desc : descriptors) {
    if(desc.direction != Direction::Read) continue;

    enablePeriodicReadList.emplace_back(ChimeraTK::ScalarPollInput<int>(this, desc.pvName + "_enaPeriodicRead", "",
                                            "Enable periodic read-out of " + desc.pvName),
        desc.value);
  }
}

/**********************************************************************************************************************/

void TecModule::mainLoop() {
  std::cout << "Processing TecModule::mainLoop()" << std::endl;

  tecIface = std::make_shared<TecChimeraInterface>(
      triggerCommands, responseStatus, commandMemoryRaw_wr, commandMemoryRaw_rd, true);
  tec = TecController(tecIface);

  /* This code is just for demonstration/testing purposes. We can use code like this to revese engineer the teclib
   * when implementing the TecDummy backend. It will be removed later! */
  tec.addFrame(0, TecFrame::TecParameter::ObjMaxTempChange, TecFrame::TecFrameType::TecParameterValueSet, 1, 0.0F);
  tec.addFrame(0, TecFrame::TecParameter::DeviceStatus, TecFrame::TecFrameType::TecParameterValueRead, 1);
  std::cout << "constructAllFrames:" << std::endl;
  tec.constructAllFrames();
  for(uint8_t i = 0; i < tec.getFrameCount(); i++) {
    tec.getFrameAt(i)->print();
  }
  std::cout << "copyCommandsToHardware:" << std::endl;
  tec.copyCommandsToHardware();
  std::cout << "triggerHardwareCommands:" << std::endl;
  tec.triggerHardwareCommands();
  std::cout << "readHardwareResponses:" << std::endl;
  auto ret = tec.readHardwareResponses();
  std::cout << " -> return " << ret << std::endl;
  std::cout << "copyResponsesFromHardware:" << std::endl;
  tec.copyResponsesFromHardware();
  std::cout << "parseAllFrames:" << std::endl;
  tec.parseAllFrames();
  for(uint8_t i = 0; i < tec.getFrameCount(); i++) {
    std::cout << "STATE = " << tec.getFrameAt(i)->getState() << std::endl;
    tec.getFrameAt(i)->print();
  }
  std::cout << "done." << std::endl;

  // Create TransferElementID to TecParameter maps
  createCommandGroupMaps();

//Debug output for TecParameter maps
//  for(auto entry : readGroupMap){
//      std::cout << "Read parameters for ID: " << entry.first << std::endl;
//      for(auto elem : entry.second){
//        std::cout << "  "<< elem << std::endl;
//      }
//  }
//  for(auto entry : writeGroupMap){
//      std::cout << "Wrt parameters for ID: " << entry.first << std::endl;
//      for(auto elem : entry.second){
//        std::cout << "  "<< elem << std::endl;
//      }
//  }
//  for(auto entry : writeGroupToFlashMap){
//      std::cout << "Wrt to flash parameters for ID: " << entry.first << std::endl;
//      for(auto elem : entry.second){
//        std::cout << "  "<< elem << std::endl;
//      }
//  }
      
  auto group = readAnyGroup();
  while(true) {
    auto id = group.readAny();
    if(id == periodicRead.trigger.getId()) {
      readPeriodic();
    }
    else if(readGroupMap.find(id) != readGroupMap.end()) {
      readGroup(readGroupMap.at(id));
    }
    else if(writeGroupMap.find(id) != writeGroupMap.end()) {
      writeGroup(writeGroupMap.at(id));
    }
    else if(writeGroupToFlashMap.find(id) != writeGroupToFlashMap.end()) {
      writeGroupToFlash(writeGroupToFlashMap.at(id));
    }
    else {
      std::cout << "Unexpected input (id = " << id << ") has been written, don't know what to do!" << std::endl;
    }
  }
}

/**********************************************************************************************************************/

void TecModule::readPeriodic() {
  std::cout << "readPeroidic()" << std::endl;
  tec.clearAllFrames();
  for(auto& entry : enablePeriodicReadList) {
    if(entry.first != 0) {
      std::cout << "Querying value for " << entry.second << std::endl;
      tec.addFrame(0, entry.second, TecFrame::TecFrameType::TecParameterValueRead, 1);
    }
  }

  // No frames added, nothing to do
  if(tec.getFrameCount() == 0) return;

  tec.constructAllFrames();
  tec.copyCommandsToHardware();
  tec.triggerHardwareCommands();

  while(not tec.readHardwareResponses()) {
    // TODO: Add proper timeout
    usleep(1000000);
  }

  tec.copyResponsesFromHardware();
  tec.parseAllFrames();

  for(size_t i = 0; i < tec.getFrameCount(); i++) {
    auto frame = tec.getFrameAt(i);
    if(frame->getState() != TecFrame::TecFrameParsed) {
      std::cout << "Error: Frame not parsed, skipping" << std::endl;
      continue;
    }

    try {
      auto& accessor = readIntMap.at(frame->getParameterOffset());
      accessor = frame->getIntParameterValue();
      accessor.write();

      continue;
    } catch(std::out_of_range&)  {
      // Intentionally left blank, could be float
    }

    try {
      auto& accessor = readFloatMap.at(frame->getParameterOffset());
      accessor = frame->getFloatParameterValue();
      accessor.write();
    } catch(std::out_of_range&)  {
      // FIXME: Once done with mapping, throw exception?
      std::cout << "Unmapped TecParameter " << frame->getParameterOffset() << ", skipping" << std::endl;
    }
  }
}

/**********************************************************************************************************************/

void TecModule::readGroup(std::list<TecFrame::TecParameter>& group) {
  std::cout << "readGroup()" << std::endl;
  for(auto& param : group) std::cout << "Read param: " << param << std::endl;
}

/**********************************************************************************************************************/

void TecModule::writeGroup(std::list<TecFrame::TecParameter>& group) {
  std::cout << "writeGroup()" << std::endl;
  tec.clearAllFrames();

  for(auto& param : group) {
    try {
      auto& accessor = writeIntMap.at(param);
      TecFrame frame{0, param, TecFrame::TecFrameType::TecParameterValueSet, 1, accessor};
      tec.addFrame(frame);
    }
    catch(std::out_of_range& ex) {
      // FIXME: Once done, throw exception?
      std::cout << "Unmapped TecParameter " << param << ", skipping" << std::endl;
    }
  }

  tec.constructAllFrames();
  tec.copyCommandsToHardware();
  tec.triggerHardwareCommands();

  while(not tec.readHardwareResponses()) {
    // TODO: Add proper timeout
    usleep(1000000);
  }

  tec.copyResponsesFromHardware();
  tec.parseAllFrames();

  for(auto i = 0; i < tec.getFrameCount(); i++) {
    auto f = tec.getFrameAt(i);
    if(f->getState() != TecFrame::TecFrameState::TecFrameParsed) continue;

    try {
      auto& accessor = readIntMap.at(f->getParameterOffset());
      accessor = f->getIntParameterValue();
      accessor.write();
    } catch(std::out_of_range&)  {
      // FIXME: Once done with mapping, throw exception?
      std::cout << "Unmapped TecParameter " << f->getParameterOffset() << ", skipping" << std::endl;
    }
  }
}

/**********************************************************************************************************************/

void TecModule::writeGroupToFlash(std::list<TecFrame::TecParameter>& group) {
  std::cout << "writeGroupToFlash()" << std::endl;
  for(auto& param : group) std::cout << "Write param to flash: " << param << std::endl;
}

/**********************************************************************************************************************/

/**
 * Iterates over all ParameterDescriptors and maps the TransferElementID of
 * the related input in groupCommandMap to the TecParameter value
 */
void TecModule::createCommandGroupMaps() {
  for(auto descriptor : descriptors){
    if(!descriptor.readGroup.empty()){
      readGroupMap[groupCommandMap.at(descriptor.readGroup).getId()].emplace_back(descriptor.value);
    }
    if(!descriptor.writeGroup.empty()){
      writeGroupMap[groupCommandMap.at(descriptor.writeGroup).getId()].emplace_back(descriptor.value);
    }
    if(!descriptor.writeToFlashGroup.empty()){
      writeGroupToFlashMap[groupCommandMap.at(descriptor.writeToFlashGroup).getId()].emplace_back(descriptor.value);
    }
  }
}

/**********************************************************************************************************************/
